/*----------------------------------------------------------------------------
;
; phcom -- command line COM port I/O.  (c) 2017 PaulHoule.com
;
; Also can read/write stdin/stdout, can be useful to generate test files.
;
; Compile Microsoft VC 7.1 or later, 32 or 64 bit:  cl /O1 /Ob0 phcom.c
;
; Released version was linked with wcrt.lib (ibsensoftware.com) and
; winpf.lib (paulhoule.com) and UPX'd to yield a tiny executable -- this is
; optional and has no effect on functionality.
;
;
; Change Log:
;	4/12/2017 Ver 1.0:	Initial release.
;	4/28/2017 Ver 1.1:	Only set com port state/timeouts if they are changed.
;						Fixed bug using multiple -s opts with w func, added
;						help saying -s must precede {write data} elements.
;						Added 'x' func (write followed by read).
;						Allowed -b0 to reset base to "none" (entry default).
;	3/11/2018 Ver 1.2:	Added '-d' to drop data from front of output.
;
;---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>

#ifdef PHCOM_TINY					// defined when linking with tiny run-time
  #include <winpf.h>
  #define phprintf w_printf
  #define phsprintf w_sprintf
  #define phfprintf w_fprintf
  #define phvfprintf w_vfprintf
  #define phstderr W_STDERR
#else
  #define phprintf printf
  #define phsprintf sprintf
  #define phfprintf fprintf
  #define phvfprintf vfprintf
  #define phstderr stderr
#endif

void ExitProg(struct vars *pv, unsigned e_code, ...);
void IoError(struct vars *pv, unsigned e_code);
int ParseOpt(struct vars *pv);
int ParseFunc(struct vars *pv);
int ParseData(struct vars *pv);
void ParseDataAbort(struct vars *pv);
void ParseCmdl(struct vars *pv);
void AllocBuf(struct vars *pv);
void OpenAndConfig(struct vars *pv);
void smode(struct vars *pv);
void sread(struct vars *pv);
void swrite(struct vars *pv);
void swriteread(struct vars *pv);
void RepWrite(struct vars *pv);
void bDump(struct vars *pv, unsigned sh, unsigned char *pd, unsigned len);
void bFmtDump(struct vars *pv, unsigned sh, unsigned char *pd, unsigned len);
void bFlush(struct vars *pv);

#define OPTL "cissbrrddmxtXTz"			// -opts and corresponding index
enum {	O_c, O_i, O_s, O_sH, O_b,		// (>= O_r must have arg)
		O_r, O_rH, O_d, O_dH, O_m, O_x, O_t, O_X, O_T, O_z };
#define OV_NONE	 ((int) ~0)				// option value: not specified
#define OV_EMPTY (OV_NONE - 1)			// option value: present but empty
#define O(o) pv->ov[O_##o]				// shorthand for option o value
#define OW(o) (*(unsigned long long *)&O(o))  // shorthand for wide opt value

enum { E_SUCCESS, E_HELP, E_BADARG,		// program exit codes
	E_NOFUNC, E_NOCOM = 48, E_NOMEM, E_STATE = 64, E_STTIMEOUT,
	E_READ = 80, E_RDTIMEOUT, E_WRITE = 96, E_WRTIMEOUT };

#define FUNC "mrwx"						// funcs & corresponding enum/table
enum { F_NONE, F_MODE, F_READ, F_WRITE, F_WRITEREAD };
void (*sfunc[])(struct vars *pv) = { smode, sread, swrite, swriteread };

struct vars {							// all vars
	unsigned ov[sizeof(OPTL) - 1];		// opt values (OV_NONE if unspecified)
	unsigned func;						// func to perform (see F_...)
	unsigned buflen;					// count of bytes now at *pBuf
	unsigned argi;						// next argv index to parse
	char **argv;						// unchanged (program entry) argv
	unsigned char *pBuf;				// ptr to data buffer
	HANDLE hCom;						// handle to COM port
	DCB dcb;							// Win serial control block
	COMMTIMEOUTS cto;					// Win communication timeouts
} v;

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Entry point -- steps through main program stages,
*/

int main(int argc, char **argv) {
	struct vars *pv = &v;

	pv->argv = argv;					// needed by all
	ParseCmdl(pv);						// parse command line
	AllocBuf(pv);						// allocate data buffer (pv->pBuf)
	OpenAndConfig(pv);					// open and configure the COM port
	sfunc[pv->func - 1](pv);			// perform selected function
	ExitProg(pv, E_SUCCESS);
	return 0;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Displays COM port mode (phcom m).  Called from main via sfunc table.
*/

void smode(struct vars *pv) {
	static char *stop[] = {"1", "1.5", "2"};

	if (!pv->dcb.DCBlength) return;		// do nothing if not COM port handle
	phprintf("-c%u -m\"baud=%u parity=%c data=%u stop=%s\""
	  " -i%u -t%u -x%u -T%u -X%u\n",
	  O(c), pv->dcb.BaudRate, "noems"[pv->dcb.Parity],
	  pv->dcb.ByteSize, stop[pv->dcb.StopBits],
	  pv->cto.ReadIntervalTimeout, pv->cto.ReadTotalTimeoutConstant,
	  pv->cto.ReadTotalTimeoutMultiplier, pv->cto.WriteTotalTimeoutConstant,
	  pv->cto.WriteTotalTimeoutMultiplier
	);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Reads from COM port (phcom r).  Called from main via sfunc table.
; By default reads till timeout.
;
;	-s#: limit read to # bytes.  Note this is always active; if user hasn't
;		 specified one the limit becomes 2**64 (effectively infinite).
*/

void sread(struct vars *pv) {
	DWORD bio;
	unsigned j;

	while (1) {
		j = O(z);						// assume read of entire buffer
		if (j > OW(s)) {				// enforce -s
			j = O(s);
			if (!j) return;				// done if read limit hit
		}

		if (!ReadFile(pv->hCom, pv->pBuf, j, &bio, NULL)) break;

		bFmtDump(pv,					// dump data using -b format control
		  STD_OUTPUT_HANDLE, pv->pBuf, bio);

		if (!bio) return;				// zero bytes read is timeout/EOF
		OW(s) -= bio;					// update -s
		if (j == bio) continue;			// if got full request, read more

		if (pv->dcb.DCBlength) return;	// done -- COM port timed out
	}
	if (GetLastError() == ERROR_BROKEN_PIPE) return; // hit pipe EOF, normal
	IoError(pv, E_READ);				// else abort I/O error
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Writes to COM port (phcom w).  Called from main via sfunc table.
;
;	-s#: limit write to # bytes (see sread).
*/

void swrite(struct vars *pv) {
	pv->buflen = 0;						// empty buffer (clear prescan mode)
	OW(r) = OV_NONE;					// clear potential prescanned -r
	OW(d) = OV_NONE;					// clear potential prescanned -d
	while (1) {
		unsigned long long d = OW(d);	// save to test for -d change
		int i;
		if (!pv->argv[++pv->argi]) break; // break= hit end of arg list
		i = ParseOpt(pv);				// try to parse arg as option
		if (i >= 0) {					// if option parsed succesfully,
			if (i == O_d) {				// if -d option parsed,
				unsigned long long nd = OW(d);	// save new D value
				OW(d) = d;				// restore previous -d
				bFlush(pv);				// flush old data per old -d
				OW(d) = nd;				// restore new -d value
				continue;
			}
			if (i == O_s) {				// if -s option parsed,
				OW(s) += pv->buflen;	// flush buffer (w/o changing new -s)
				bFlush(pv);
			}
			continue;
		}

		if (OW(r) == OV_NONE) {			// typical case is OV_NONE
			ParseDataAbort(pv);			// process data arg, no repeat
		} else {						// else arg preceeded by -r#,
			RepWrite(pv);				// handle with optimized repeat logic
		}
	}
	bFlush(pv);							// flush any data left in buffer
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Writes to COM port, then reads from COM port (phcom x).
;
;	-s#: a final -s on the command line will limit the read size.
*/

void swriteread(struct vars *pv) {
	OW(s) = OV_NONE;					// clear ending (read size) -s
	swrite(pv);							// first write data
	if (O(c) == OV_EMPTY) {				// if empty -c, read from stdin
		pv->hCom = GetStdHandle(STD_INPUT_HANDLE);
	}
	sread(pv);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Allocates main data buffer -- user may specify buffer size via -z option.
; If -z unspecified or 0 a reasonable default size (see code below) is used.
;
; Primarily for debug; buffer size doesn't matter as long as it's not so tiny
; that performance suffers.  In one obscure case it might -- if a pipe is used
; for write data and repeated (via -r) the repeat will be ignored unless the
; entire pipe fits in the buffer (since a pipe can't be re-read).  If that
; obscure case is anticipated the user may wish to set a large buffer size.
*/

void AllocBuf(struct vars *pv) {
	if (!O(z) || O(z) == OV_NONE) O(z) = 16384;
	pv->pBuf = malloc(O(z));
	if (!pv->pBuf) ExitProg(pv, E_NOMEM, "Insufficient memory\n");
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; RepWrite/rwCopy are used when writing with an active -r#.  Extra logic is
; applied to optimize buffering (eg speed up "-r9999999999 ff").
;
; Note: -r count is 64-bits.
*/

void _fastcall rwCopy(char *pDst, char *pSrc, unsigned c) {
	unsigned cnt = c;					// (otherwise compiler won't register)
	if (cnt) do { *pDst++ = *pSrc++; } while (--cnt);
}
void RepWrite(struct vars *pv) {
	unsigned i, j;

	while (1) {
		bFlush(pv);						// empty to have full use of buffer
		if (OW(r)) {					// if at least one copy to do,
			unsigned long long s = OW(s); // save for "fit in buf" check
			ParseDataAbort(pv);			// generate arg at start of buf
			if (s != OW(s)) {			// if didn't fit (flush was done),
				--OW(r);				// can't optimize, do one at a time
				continue;
			}
		}

		j = pv->buflen;					// j= byte size of parsed arg
		if (!j) break;					// break= nothing to repeat
		i = O(z) / j;					// i= total copies that fit (1..n)
		if (i > OW(r)) i = O(r);		// limit to what we need
		rwCopy(pv->pBuf + j, pv->pBuf, j * (i - 1)); // generate copies
		j *= i;							// j= total bytes (all copies)

		while (1) {
			if (j >= OW(s)) {			// if stopped by -s size limit,
				j = O(s);				// exit w/residual in buf
				break;
			}
			if (i >= OW(r)) {			// if have all copies we need,
				j /= i;					// (potentially) downsize buflen to
				j *= O(r);				//	size of copies we need
				break;					// done, exit w/residual in buf
			}
			OW(r) -= i;
			pv->buflen = j;				// flush j bytes
			bFlush(pv);
		}
		pv->buflen = j;					// update buffered data amount (0..n)
		break;
	}
	OW(r) = OV_NONE;					// clear -r
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Implements drop (-d option) on len, returning len.  If no -d active, len
; is unchanged.  Note: a drop may zero the returned len value.
*/

unsigned _fastcall DoDrop(struct vars *pv, unsigned len) {
	if (OW(d) != OV_NONE) {				// if -d active,
		if (OW(d) >= len) {				// if drop exceeds buffered data,
			OW(d) -= len;				// simply reduce drop amount
			return 0;
		}
		len -= O(d);					// -d completed, reduce len
		OW(d) = OV_NONE;				// terminate -d mode
	}
	return len;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Writes len bytes from *pd to pv->hCom (or STD_... handle if sh != 0).
;
; Aborts if timeout/hard error.
*/

void bDump(struct vars *pv, unsigned sh, unsigned char *pd, unsigned len) {
	HANDLE h = pv->hCom;				// assume we'll use COM port handle
	DWORD bio;
	unsigned i = E_WRITE;				// init to hard error status

	if (!len) return;					// don't bother writing zero bytes
	if (sh) h = GetStdHandle(sh);		// handle STD_... case

	// We assume a single write MUST write all data or it's an error.
	if (WriteFile(h, pd, len, &bio, NULL)) {
		if (len == bio) return;			// if all written, we're good

		if (h == pv->hCom) {			// if using our configured handle,
			if (pv->dcb.DCBlength)		// if handle sensed to be COM port,
			  i = E_WRTIMEOUT;			// set timeout status
		}
	}
	ExitProg(pv, i,						// abort due to timeout/hard error
	  "Write %s\n", i == E_WRTIMEOUT ? "Timeout" : "Error");
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Wrapper around bDump -- bytes are written as specified by the -b switch.
; If no -b specified a binary dump is performed; otherwise, the data is
; dumped as space separated numbers of base O(b).

; Aborts if timeout/hard error.
*/

void bFmtDump(struct vars *pv, unsigned sh, unsigned char *pd, unsigned len) {
	char nbuf[56 * 9];					// 56 worst case (binary + space)
	unsigned olen = len;

	len = DoDrop(pv, len);				// apply -d (if one pending)
	if (olen -= len) pd += olen;		// if dropped, advance ptr
	if (O(b) == OV_NONE) {				// default is ascii output
		bDump(pv, sh, pd, len);
		return;
	}

	while (1) {
		char *pw = nbuf + sizeof(nbuf);	// init number storage ptr
		unsigned i;
		unsigned j = sizeof(nbuf) / 9;	// j= elements nbuf will hold
		if (O(b) >= 8) j = sizeof(nbuf) / 4; // (either 56 or 126)
		if (j > len) {					// if near/at end
			j = len;					// limit to what's left to do
			if (!j) break;				// break= nothing left
		}
		i = j;							// save for adjusting pd/len
		do {
			unsigned k = pd[j - 1];		// format number to specified base
			*--pw = ' ';				// (always place a trailing space)
			do {
				char c = (char) ('0' + k % O(b));
				if (c > '9') c += 'a' - ('9' + 1);
				*--pw = c;
			} while (k /= O(b));
		} while (--j);
		pd += i;						// update source read ptr/len
		len -= i;
		bDump(pv, sh, pw, (unsigned) (&nbuf[sizeof(nbuf)] - pw));
	}
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Opens COM port and assigns handle of opened port.
;
; By default scans down from 30 for first openable port and sets port #
; in -c value (finds USB<->Serial dongles which are assigned high port).
;
; If -c0 used, scans down from 256 instead of 30.
;
; If a port can't be opened aborts w/error.
*/

void OpenCom(struct vars *pv) {
	unsigned i = 30;					// assume backscanning from 30

	if (O(c) < OV_EMPTY) {				// if -c specified,
		i = O(c);						// load port
		if (!i) i = 256;				// zero means backscan from 256
	}
	i++;								// compensate for predecrement
	do {
		char namebuf[16];
		phsprintf(namebuf, "\\\\.\\COM%u", --i);
		pv->hCom = CreateFile(namebuf, GENERIC_READ | GENERIC_WRITE, 0,
			NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
		if (pv->hCom != INVALID_HANDLE_VALUE) {
			O(c) = i;					// store COM # (in case scanning)
			return;
		}
	} while (i != O(c) && i > 0);
	ExitProg(pv, E_NOCOM, "Can't open COM, last port tried %u\n", i);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Gets then Sets (if needed) COM state (baud, parity, data/stop bits, etc).
; If couldn't get state we assume handle isn't a COM port, do nothing
; further, and zero pv->dcb.DCBlength to indicate "non-COM" handle.
;
; If failure during get/set aborts w/error.
*/

void SetComState(struct vars *pv) {
	pv->dcb.DCBlength = sizeof(pv->dcb);
	if (!GetCommState(pv->hCom, &pv->dcb)) {
		pv->dcb.DCBlength = 0;			// indicate hCom not a COM port
		return;
	}
	do {
		if (O(m) == OV_NONE) return;	// don't set if -m not specified
		if (!BuildCommDCB(				// build the dcb
			pv->argv[O(m)] + 2, &pv->dcb)) break;
		pv->dcb.fAbortOnError = FALSE;	// necessary to prevent port hang?
		if (SetCommState(pv->hCom, &pv->dcb)) return;
	} while (0);
	ExitProg(pv, E_STATE, "Error setting COM state\n");
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Gets then Sets (if needed) COM timeouts.  SetComState must have been called.
;
; If failure during get/set aborts w/error.
*/

void SetComTimeouts(struct vars *pv) {
	COMMTIMEOUTS *pcto = &pv->cto;
	if (!pv->dcb.DCBlength) return;		// do nothing if not COM port handle
	do {
		unsigned i;
		if (!GetCommTimeouts(pv->hCom, pcto)) break;
		i = 0;							// 0 indicates we've made no change
		if (O(i) != OV_NONE) i++, pcto->ReadIntervalTimeout =
			O(i) == OV_EMPTY ? MAXDWORD : O(i);
		if (O(x) != OV_NONE) i++, pcto->ReadTotalTimeoutMultiplier = O(x);
		if (O(t) != OV_NONE) i++, pcto->ReadTotalTimeoutConstant = O(t);
		if (O(X) != OV_NONE) i++, pcto->WriteTotalTimeoutMultiplier = O(X);
		if (O(T) != OV_NONE) i++, pcto->WriteTotalTimeoutConstant = O(T);
		if (!i) return;
		if (SetCommTimeouts(pv->hCom, pcto)) return;
	} while (0);
	ExitProg(pv, E_STTIMEOUT, "Error getting/setting COM timeouts\n");
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Opens and configures I/O -- setting pv-hCom, pv->dcb, pv->cto.
; If we scanned for it, also sets O(c) to the true COM port #.
;
; Typically we open a COM port, though there is the empty "-c" case to
; force use of stdin/stdout; then the active handle can be anything
; (eg. file, pipe, CON, or even COMx).
;
; If failure aborts w/error.
*/

void OpenAndConfig(struct vars *pv) {
	if (O(c) == OV_EMPTY) {				// if empty -c, use stdin/stdout
		pv->hCom = GetStdHandle(
			pv->func == F_READ ? STD_INPUT_HANDLE : STD_OUTPUT_HANDLE);
	} else {
		OpenCom(pv);
	}
	SetComState(pv);					// get/set state (eg baud rate, etc)
	SetComTimeouts(pv);					// get/set timeouts
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Reports unexpected I/O error and aborts returning specified e_code.
*/

void IoError(struct vars *pv, unsigned e_code) {
	ExitProg(pv, e_code, "I/O error %u\n", GetLastError());
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; General purpose release resources and task exit.
; e_code= non-zero if error -- fprintf args must follow.
*/

void ExitProg(struct vars *pv, unsigned e_code, ...) {
	if (O(c) != OV_EMPTY) CloseHandle(pv->hCom); // if did open, do close
	free(pv->pBuf);						// free data buffer
	if (e_code) {						// if error, display message
		char *pFmt;
		va_list vl;
		va_start(vl, e_code);
		pFmt = va_arg(vl, char *);
		if (pFmt) {
			phfprintf(phstderr, "phcom: ");
			phvfprintf(phstderr, pFmt, vl);
		}
		va_end(vl);
	}
	exit(e_code);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Help text.
; In <1k pieces due to limits of the Win user32.lib wvsprintf() we might use.
*/

char htxt0[] = {
	"\nCOM port I/O, ver 1.2.  (c) paulhoule.com 3/11/2018\n\n"
	"  Syntax:  phcom {-opts} func {write data}\n"
	"\nfunc:\n"
	"  m (mode)  displays port used and its settings\n"
	"  r (read)  reads data from COM port\n"
	"  w (write) writes {write data} to COM port\n"
	"  x (w+r)   writes then reads, final -s specifies read amount\n"
	"\n-opts (-s, -r, -d are 64-bit values):\n"
	"  -c#   Force COM port # (no -c scans down from 30, -c0 from 256)\n"
	"        [-c alone I/O's stdin/stdout instead of COM port]\n"
	"  -m$   Sets COM mode to string $ (same format as MODE COMx command)\n"
	"  -i#   read interval timeout ms (empty value sets to MAXDWORD)\n"
};
char htxt1[] = {
	"  -x/X# read(x)/write(X) timeout per-byte multiplier ms\n"
	"  -t/T# read(t)/write(T) timeout constant ms\n"
	"  -s#   Limit, but not increase, output size\n"
	"  -r#   repeat following data arg # times\n"
	"  -d#   drop # bytes from subsequent output\n"
	"  -b#   sets base of write data (-b is hex, -b0 resets to none);\n"
	"        other than -b0 causes read output to be numeric\n"
	"  -z#   debug: set data buffer size to # (1..n, default is 16k)\n"
	"\n{write data}:\n"
	"   hex bytes (by default, see -b), or single-quoted ascii strings,\n"
	"   or \\\"filename\\\" (filename empty for stdin)\n"
};

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Display help and exit.
*/

void HelpExit(struct vars *pv) {
	phprintf(htxt0);					// give help (in 1k max pieces)
	phprintf(htxt1);
	ExitProg(pv, E_HELP, NULL);			// exit non-zero, no error msg
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Converts number to unsigned if llf == 0, else to unsigned long long.
; When llf == 0, high unsigned of result should be ignored.
;
; If error/overflow result is (unsigned)/(unsigned long long) -1.
*/

unsigned _fastcall hxld(unsigned *pl, unsigned h, unsigned long long d) {
	unsigned long long i = (unsigned long long)h * *pl + d;
	*pl = (unsigned) i;
	return (unsigned) (i >> 32);
}
unsigned long long strtoux(char *pArg, char **pE, unsigned base, int llf) {
	unsigned long long r = 0;
	if (base <= 16 && *pArg == '0' && ((pArg[1] | 0x20) == 'x')) {
		pArg += 2, base = 16;			// allow 0x override if base <= 16
	}
	*pE = pArg;
	do {
		unsigned d = (unsigned char) *pArg;
		if (d < '0') break;				// break= non-digit
		if (d > '9') {
			d |= 0x20;					// convert to lowercase
			if (d < 'a') break;			// break= non-digit
			d -= 'a' - ('9' + 1);		// convert 'a'... to '9'+1...
		}
		d -= '0';
		if (d >= base) break;			// break= non-digit
		pArg++;							// absorb digit
		d = hxld(0 + (unsigned *)&r, base, d);
		if (llf) d = hxld(1 + (unsigned *)&r, base, d);
		if (d) return -1;
	} while (*pArg);
	if (pArg == *pE) return -1;			// abort if no digit processed

	*pE = pArg;							// return ptr to terminator
	return r;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Flushes pv->buflen bytes of write data starting at *pv->pBuf.
; Reduces -s counter, pv->buflen is reset to zero.
;
; Note: this routine may not alter the contents of the data buffer.
; Note: pv->buflen (except for -1 "prescan" case) must never exceed OW(s).
;
; Aborts program if write error or timeout.
*/

void bFlush(struct vars *pv) {
	unsigned j = pv->buflen;
	unsigned char *pBuf = pv->pBuf;		// local ptr to data buffer

	OW(s) -= j;							// update -s
	j = DoDrop(pv, j);					// apply -d (if one pending)
	if (pv->buflen - j) pBuf += pv->buflen - j;	// if dropped, advance ptr
	bDump(pv,							// flush buffer
	  O(c) == OV_EMPTY ? STD_OUTPUT_HANDLE : 0, pBuf, j);
	pv->buflen = 0;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Returns buffer size limit that respects -s.
; Will also flush the buffer if it is completely full.
*/

unsigned CalcLim(struct vars *pv) {
	unsigned lim = O(z);				// assume limit is end-of-buffer
	if (lim == pv->buflen) {			// if buffer is full, flush it
		bFlush(pv);
	}
	if (lim > OW(s)) lim = O(s);		// enforce -s
	return lim;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Parses file data arg.  pFn points to filename (after lead double quote).
; Note: pv->buflen == -1 is a prescan "don't read data" mode.
;
; Return:	ptr past filename end double quote= success.
;			NULL= error opening file (prob doesn't exist -- bad arg).
;
; Program aborted if error reading file data.
*/

char *ParseFileData(struct vars *pv, char *pFn) {
	HANDLE h;
	DWORD bio;
	unsigned i;
	char *pE = pFn;
	char lc;

	do { lc = *pE++; } while (lc && lc != '"'); // find end of filename
	*--pE = '\0';						// terminate (we'll repair later)

	if (pFn == pE) {					// empty name reads from stdin
		h = GetStdHandle(STD_INPUT_HANDLE);
	} else {
		h = CreateFile(pFn, GENERIC_READ, FILE_SHARE_READ,
			NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
	}

	if (h != INVALID_HANDLE_VALUE) {	// if got valid handle,
		if (pv->buflen != -1) {			// only read if not prescan mode
			while (1) {
				unsigned lim;
				lim = CalcLim(pv);		// lim= max allowed buflen
				lim -= pv->buflen;		// convert to free buffer space
				if (!lim) break;		// break= done (hit -s limit)
				i = ReadFile(h, pv->pBuf + pv->buflen, lim, &bio, NULL);
				if (!i && GetLastError() != ERROR_BROKEN_PIPE)
					IoError(pv, E_READ); // abort= read hard error
				pv->buflen += bio;
				if (!bio) break;		// break= hit eof
			}
		}
		if (pFn != pE) CloseHandle(h);	// close handle if we opened it
	}

	if (lc) *pE++ = lc;					// repair arg, skip a closing quote

	if (h == INVALID_HANDLE_VALUE) pE = NULL; // return NULL if open error
	return pE;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Parses data arg at index pv->argi.
; Note: pv->buflen == -1 is a prescan "don't store data" mode.
;
; Return:	 0= success.
;			-1= bad arg (improper number, or file couldn't be opened).
*/

int ParseData(struct vars *pv) {
	char *pArg = pv->argv[pv->argi];
	char *ptmp;
	unsigned i, j, k;
	unsigned lim = 0;					// indicate lim needs to be recalc'd

	j = 0;								// string flag: 1=inside, 0=outside
	do {
		i = (unsigned char) *pArg++;	// fetch next byte, advance
		if ((char) i == '\'') {			// if string boundary char,
			if (j ^= 1) continue;		// ignore if entering string

			if (*pArg != '\'') continue; // ignore if exiting string
			j++;						// doubled, treat as literal char
			pArg++;
		}
		if (!j) {						// if not within string,
			if ((char) i == ' ') continue; // ignore leading whitepace
			if ((char) i == '"') {		// handle read file case
				pArg = ParseFileData(pv, pArg);
				if (!pArg) return -1;	// abort= couldn't open file
				lim = 0;				// indicate lim needs to be recalc'd
				continue;
			}
			k = (unsigned) strtoux(		// convert to unsigned
			  --pArg, &ptmp, O(b) == OV_NONE ? 16 : O(b), 0);
			i = (unsigned char) k;
			if (i != k) return -1;		// error if wider than a byte
			pArg = ptmp;				// use ptmp so pArg gets registered
		}
		if (pv->buflen >= lim) {		// if recalc/prescan/hit-limit,
			if (pv->buflen == -1) continue;	// continue (no store) if prescan
			lim = CalcLim(pv);			// lim= max allowed buflen
			if (pv->buflen == lim) break; // break= hit -s limit, ignore data
		}
		pv->pBuf[pv->buflen++] = (char) i;
	} while (*pArg);
	return 0;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Aborts prog with bad argument message.
*/

void ArgAbort(struct vars *pv) {
	ExitProg(pv, E_BADARG, "Error with arg: %s\n", pv->argv[pv->argi]);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Same as ParseData except aborts w/bad-arg message if error occurs.
*/

void ParseDataAbort(struct vars *pv) {
	if (ParseData(pv)) ArgAbort(pv);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Parse suspected option at pv->argv[pv->argi].
; Some haphazard option validation is done.
;
; All options stored unsigned except O_s/O_r/O_d which are stored as
; unsigned long long (they take two adjacent unsigned pv->ov[] slots).
;
; Return: O_... enum of parsed opiton (0..n)= success, -1= error.
*/

int ParseOpt(struct vars *pv) {
	static const char opts[] = OPTL;
	char *pArg = pv->argv[pv->argi];
	const char *pO = opts - 1;			// ptr to sequence through opts
	int i;
	unsigned long long v;				// option value

	if (*pArg != '-' && *pArg != '/') return -1;
	pArg++;
	if (*pArg == '?') HelpExit(pv);
	do {
		if (!*++pO) return -1;
	} while (*pO != *pArg);
	pArg++;

	if (*pO == 'm') {					// special case port mode string
		O(m) = pv->argi;				// save option argv index
		return O_m;						// success
	}

	if (*pArg) {						// if value to parse,
		char *pE;
		v = strtoux(pArg, &pE, 10, 1);	// convert to unsigned long long
		if (v >= OV_EMPTY) return -1;	// this is bad regardless of width
		if (*pO != pO[1]) {				// if unsigned width option,
			if (v >= (unsigned) OV_EMPTY) return -1; // must fit in unsigned
		}
		if (*pE) return -1;				// error if bad termination
	} else {							// else no arg present
		v = OV_EMPTY;
		if (pO >= &opts[O_r]) return -1; // error if arg required
	}
	if (*pO == 'b') {
		if ((unsigned) v == OV_EMPTY) v = 16; // force -b to be same as -b16
		if ((unsigned) v == 0) {
			v = OV_NONE;				// -b0 resets to none (ascii output)
		} else {
			if ((unsigned) v < 2 || (unsigned) v > 36) return -1;
		}
	}

	i = (int) (pO - opts);				// option pv->ov[] index
	pv->ov[i] = (unsigned) v;			// store low unsigned
	if (*pO == pO[1]) {					// if double unsigned width option,
		pv->ov[i + 1] = (unsigned) (v >> 32); // store high unsigned
	}
	return i;							// success, return option index 0..n
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Parse func arg at index pv->argi.
;
; Return: 0= success, -1= error.
*/

int ParseFunc(struct vars *pv) {
	char *pArg = pv->argv[pv->argi];
	unsigned i = 0;						// i= func index (F_... enum)

	do { if (!FUNC[i]) return -1; } while (FUNC[i++] != *pArg); // find func
	if (pArg[1]) return -1;				// func code must be single letter

	pv->func = i;						// set func code F_ value
	return 0;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Parses entire command line -- func code and all options.
; {data} fields (if any) are parsed but skipped.
;
; In:	pv= ptr to allocated vars structure
;		pv->argv= main argv vector, must be terminated w/NULL ptr
;
; Out:	initialized -- pv->buflen, pv->argi, pv->ov, pv->func
;		pv->argi left indexing func arg for subsequent arg reparsing
;
; Unspecified options are set to OV_NONE; specified option with no value
; set to OV_EMPTY.  Options are unsigned (neg options not implemented).
;
; If parsing error or empty line displays error/help and aborts.
*/

void ParseCmdl(struct vars *pv) {
	unsigned i;

	i = sizeof(pv->ov) / sizeof(*pv->ov);			// init option values
	do { pv->ov[i - 1] = OV_NONE; } while (--i);	// to OV_NONE

	pv->func = 0;						// indicate func not yet parsed
	pv->buflen = -1;					// set prescan mode for ParseData
	pv->argi = 0;						// init arg index
	while (1) {
		if (!pv->argv[++pv->argi]) {	// if hit end of arg list,
			pv->argi = i;				// reset index to func arg
			if (pv->func) return;
			HelpExit(pv);
		}
		if (ParseOpt(pv) >= 0) continue; // continue= parsed -opt succesfully

		if (!pv->func) {				// if func not yet parsed,
			if (ParseFunc(pv)) break;	// break= neither opt or func
			i = pv->argi;				// save index of func arg
			continue;					// contine= parsed as func
		}
		if (pv->func != F_WRITE &&		// break= err, only w/x can have args
			pv->func != F_WRITEREAD) break;
		if (ParseData(pv)) break;		// parse data arg, w/o processing it
	}
	ArgAbort(pv);
}
